from mastobot import Bot, Favourite
from os import environ

instance = environ.get("INSTANCE")
token = environ.get("ACCESS_TOKEN")

bot = Bot(instance, token, websocket_mode=True)


@bot.on_home_update(":(cis|trans|ultra)?cate:", validation="regex")
def cated(status):
    return Favourite

bot.run()



"""
https://github.com/TeamHG-Memex/html-text/blob/master/html_text/html_text.py

MIT License

Copyright (c) 2016, Konstantin Lopukhin

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

This script is modified for Mastobot.
"""

# -*- coding: utf-8 -*-
import re

import lxml
import lxml.etree
from lxml.html.clean import Cleaner


NEWLINE_TAGS = frozenset(
    [
        "article",
        "aside",
        "br",
        "dd",
        "details",
        "div",
        "dt",
        "fieldset",
        "figcaption",
        "footer",
        "form",
        "header",
        "hr",
        "legend",
        "li",
        "main",
        "nav",
        "table",
        "tr",
    ]
)
DOUBLE_NEWLINE_TAGS = frozenset(
    [
        "blockquote",
        "dl",
        "figure",
        "h1",
        "h2",
        "h3",
        "h4",
        "h5",
        "h6",
        "ol",
        "p",
        "pre",
        "title",
        "ul",
    ]
)

cleaner = Cleaner(
    scripts=True,
    javascript=False,  # onclick attributes are fine
    comments=True,
    style=True,
    links=True,
    meta=True,
    page_structure=False,  # <title> may be nice to have
    processing_instructions=True,
    embedded=True,
    frames=True,
    forms=False,  # keep forms
    annoying_tags=False,
    remove_unknown_tags=False,
    safe_attrs_only=False,
)


def _cleaned_html_tree(html):
    if isinstance(html, lxml.html.HtmlElement):
        tree = html
    else:
        tree = parse_html(html)

    # we need this as https://bugs.launchpad.net/lxml/+bug/1838497
    try:
        cleaned = cleaner.clean_html(tree)
    except AssertionError:
        cleaned = tree

    return cleaned


def parse_html(html):
    """ Create an lxml.html.HtmlElement from a string with html.
    XXX: mostly copy-pasted from parsel.selector.create_root_node
    """
    body = html.strip().replace("\x00", "").encode("utf8") or b"<html/>"
    parser = lxml.html.HTMLParser(recover=True, encoding="utf8")
    root = lxml.etree.fromstring(body, parser=parser)
    if root is None:
        root = lxml.etree.fromstring(b"<html/>", parser=parser)
    return root


_whitespace = re.compile(r"\s+")
_has_trailing_whitespace = re.compile(r"\s$").search
_has_punct_after = re.compile(r'^[,:;.!?")]').search
_has_open_bracket_before = re.compile(r"\($").search


def _normalize_whitespace(text):
    return _whitespace.sub(" ", text.strip())


def etree_to_text(
    tree,
    guess_punct_space=True,
    guess_layout=True,
    newline_tags=NEWLINE_TAGS,
    double_newline_tags=DOUBLE_NEWLINE_TAGS,
):
    """
    Convert a html tree to text. Tree should be cleaned with
    ``html_text.html_text.cleaner.clean_html`` before passing to this
    function.
    See html_text.html_to_text docstring for description of the
    approach and options.
    """
    chunks = []

    _NEWLINE = object()
    _DOUBLE_NEWLINE = object()

    class Context:
        """ workaround for missing `nonlocal` in Python 2 """

        # _NEWLINE, _DOUBLE_NEWLINE or content of the previous chunk (str)
        prev = _DOUBLE_NEWLINE

    def should_add_space(text, prev):
        """ Return True if extra whitespace should be added before text """
        if prev in ("@", "#"):
            return False
        elif str(prev).endswith("://"):
            # URL scheme
            return False
        if prev in {_NEWLINE, _DOUBLE_NEWLINE}:
            return False
        if not guess_punct_space:
            return True
        if not _has_trailing_whitespace(prev):
            if _has_punct_after(text) or _has_open_bracket_before(prev):
                return False
        return True

    def get_space_between(text, prev):
        if not text:
            return " "
        return " " if should_add_space(text, prev) else ""

    def add_newlines(tag, context):
        if not guess_layout:
            return
        prev = context.prev
        if prev is _DOUBLE_NEWLINE:  # don't output more than 1 blank line
            return
        if tag in double_newline_tags:
            context.prev = _DOUBLE_NEWLINE
            chunks.append("\n" if prev is _NEWLINE else "\n\n")
        elif tag in newline_tags:
            context.prev = _NEWLINE
            if prev is not _NEWLINE:
                chunks.append("\n")

    def add_text(text_content, context):
        text = _normalize_whitespace(text_content) if text_content else ""
        if not text:
            return
        space = get_space_between(text, context.prev)
        chunks.extend([space, text])
        context.prev = text_content

    def traverse_text_fragments(tree, context, handle_tail=True):
        """ Extract text from the ``tree``: fill ``chunks`` variable """
        add_newlines(tree.tag, context)
        add_text(tree.text, context)
        for child in tree:
            traverse_text_fragments(child, context)
        add_newlines(tree.tag, context)
        if handle_tail:
            add_text(tree.tail, context)

    traverse_text_fragments(tree, context=Context(), handle_tail=False)
    return "".join(chunks).strip()


def html_to_text(
    html,
    guess_punct_space=False,
    guess_layout=True,
    newline_tags=NEWLINE_TAGS,
    double_newline_tags=DOUBLE_NEWLINE_TAGS,
):
    """
    Convert html to text, cleaning invisible content such as styles.
    Almost the same as normalize-space xpath, but this also
    adds spaces between inline elements (like <span>) which are
    often used as block elements in html markup, and adds appropriate
    newlines to make output better formatted.
    html should be a unicode string or an already parsed lxml.html element.
    ``html_text.etree_to_text`` is a lower-level function which only accepts
    an already parsed lxml.html Element, and is not doing html cleaning itself.
    When guess_punct_space is True (default), no extra whitespace is added
    for punctuation. This has a slight (around 10%) performance overhead
    and is just a heuristic.
    When guess_layout is True (default), a newline is added
    before and after ``newline_tags`` and two newlines are added before
    and after ``double_newline_tags``. This heuristic makes the extracted
    text more similar to how it is rendered in the browser.
    Default newline and double newline tags can be found in
    `html_text.NEWLINE_TAGS` and `html_text.DOUBLE_NEWLINE_TAGS`.
    """
    if html is None:
        return ""
    cleaned = _cleaned_html_tree(html)
    return etree_to_text(
        cleaned,
        guess_punct_space=guess_punct_space,
        guess_layout=guess_layout,
        newline_tags=newline_tags,
        double_newline_tags=double_newline_tags,
    )



from ..mastobot.utils import normalize_content


def test_html_text():
    assert (
        normalize_content(
            '<p><span class="h-card"><a href="https://yeet.social/@c18n" '
            'class="u-url mention">@<span>c18n</span></a></span> text '
            '<a href="https://mastodon.technology/tags/tag" '
            'class="mention hashtag" rel="tag">#<span>tag</span></a> '
            ":cate: ®️ @ someone # sometag</p>"
        )
        == "@c18n text #tag :cate: ®️ @ someone # sometag"
    )

    assert normalize_content(
        "<p>Take a look at the "
        '<a href="https://mastodon.technology/tags/Akademy2020" '
        'class="mention hashtag" rel="tag">'
        "#<span>Akademy2020</span></a>"
        "program and mark your calendars for all of the great talks we have in store! "
        '<a href="https://akademy.kde.org/2020/program" '
        'rel="nofollow noopener noreferrer" target="_blank">'
        '<span class="invisible">https://</span>'
        '<span class="">akademy.kde.org/2020/program</span>'
        '<span class="invisible"></span></a>  <br />Don’t forget to register too! '
        '<a href="https://akademy.kde.org/2020/register" '
        'rel="nofollow noopener noreferrer" target="_blank">'
        '<span class="invisible">https://</span>'
        '<span class="">akademy.kde.org/2020/register</span>'
        '<span class="invisible"></span></a></p>'
    ) == (
        "Take a look at the #Akademy2020 program and mark your calendars "
        "for all of the great talks we have in store! "
        "https://akademy.kde.org/2020/program\n"
        "Don’t forget to register too! https://akademy.kde.org/2020/register"
    )


if __name__ == "__main__":
    test_html_text()



import re
from .structs import *
from .constants import *


class Trigger:
    def __init__(
        self, event: str, validation: str, expectation, callback, case_sensitive=None
    ):
        """Describes a condition that would trigger a callback.

        :param event: (str) may be "mention", "favourite", "reblog", "follow", or "update".
        :param validation: (str) may be "equals", "contains", "regex" or "evaluate".
        :param expectation: (str or callable) string, regex string or function
            to test content of relevant status against.

            Suppose the normalized (html-to-text converted) content of the status
            received is ``content``.

            If ``validation`` is "equals", the trigger will be invoked when
            content exactly matches expectation; if it is ``contains``, when
            expectation is found anywhere in content; if it is ``regex``,
            when content regex-matches expectation; if it is ``evaluate``,
            when ``expectation(content) == True``.            

        :param callback: (callable) callback executed when self is triggered.
            Arguments passed to it depend on ``event``.

        :param case_sensitive: (bool) whether the trigger test will be case sensitive.
            Default: ``True`` if ``validation`` is "equals" or "contains", ``False`` otherwise.
        """
        if event not in EVENT_LIST:
            raise ValueError("Trigger event incorrect")
        elif validation not in VALIDATION_LIST:
            raise ValueError("Trigger validation incorrect")
        elif not callable(callback):
            raise ValueError("Trigger callback not callable")
        elif validation == EVALUATE and not callable(expectation):
            # botdev error: nonsense criterion
            raise ValueError(
                "Trigger validation is 'evaluate' but expectation is not callable"
            )

        self.event = event
        self.validation = validation
        self.expectation = expectation
        self.callback = callback
        if case_sensitive is not None:
            self.case_sensitive = case_sensitive
        else:
            if validation in [EQUALS, CONTAINS]:
                self.case_sensitive = True
            else:
                self.case_sensitive = False

    def test(self, event: str, content: str):
        """Test if self should be triggered.

        :param event: (str) event that invoked this test.
        :param content: (str) normalized status content.
        """

        if not event == self.event:
            return

        if self.validation in [EQUALS, CONTAINS] and not self.case_sensitive:
            content = content.lower()
            expectation = self.expectation.lower()
        else:
            expectation = self.expectation

        if self.validation == EQUALS:
            return content == expectation
        elif self.validation == CONTAINS:
            return content.find(self.expectation) > -1
        elif self.validation == REGEX:
            if self.case_sensitive:
                return re.search(expectation, content)
            else:
                return re.search(expectation, content, re.IGNORECASE)
        elif self.validation == EVALUATE:
            return self.expectation(content)

    def invoke(self, obj):
        event = self.event
        if event == UPDATE:
            # what did they say?
            args = [gen_status(obj)]
        elif event == MENTION:
            # what did they say?
            args = [gen_status(obj["status"])]
        elif event in [REBLOG, FAVOURITE]:
            # who boosted/favourited your post,
            # and what was your post that made them boost/fav?
            args = [gen_account(obj["account"]), gen_status(obj["status"])]
        else:
            return

        return self.callback(*args)



from mastodon import Mastodon, StreamListener
from .ws import WebsocketListener
from .html_text import html_to_text
from .structs import *
from .trigger import Trigger
from .constants import *

# related documentation: https://mastodonpy.readthedocs.io/en/stable/#


class Bot:
    def __init__(self, instance_url: str, access_token: str, websocket_mode=False):
        """Intiate a Mastodon bot.

        :param instance_url: (str) base URL for your Mastodon instance of choice,
            e.g. ``https://mastodon.technology``.
        :param access_token: (str) "Your access token" inside
            Preferences -> Development -> some application.
        :param websocket_mode: (bool) whether to use websockets for streaming
        """
        self._instance = instance_url
        self._token = access_token
        self._handle = ""  # will be like "bot@instance.tld"
        self._atname = ""  # will be like "@bot"
        self._triggers = []
        self._websocket_mode = websocket_mode
        self._check_update_triggers = lambda o: self._check_triggers(o, UPDATE)
        self._check_notification_triggers = lambda o: self._check_triggers(
            o, NOTIFICATION
        )

    def _check_triggers(self, obj: dict, stream: str):
        """Handle events from ``mastodon.StreamListener.on_update()`` if 
            ``stream=="update"``, and ``mastodon.StreamListener.on_notification()``
            if ``stream=="notification"``.
        """
        # set up expectations
        if stream == UPDATE:
            status = obj
            event = UPDATE
        elif stream == NOTIFICATION:
            status = obj["status"]
            event = obj["type"]
        else:
            return
        # if the bot is mentioned, remove the mention text from status content
        # before testing trigger
        mentioned_accts = [m["acct"] for m in status["mentions"]]
        if self._handle in mentioned_accts:
            to_be_removed = self._atname
        else:
            to_be_removed = ""
        for trig in self._triggers:
            if not trig.event == event:
                continue
            if trig.test(
                event,
                html_to_text(status["content"]).replace(to_be_removed, "", 1).strip(),
            ):
                # pass obj to trig. trig will decide which elements
                # to pass on to bot-developer-defined callback.
                reply = trig.invoke(obj)
                if reply:
                    self._respond(status, reply)

    def _respond(self, status: dict, content):
        """Reply to a status with content, boost, or favourite it.

        :param status: (dict) the status to respond to.
        :param content: When ``content`` is a string,
            simply reply with it, keeping everything else as Mastodon.py decides.
            When it is an instance of the ``mastobot.Reply`` class, all its arguments
            will be passed on to Mastodon.py. The rest are left in their default state.
            When it is ``mastobot.Boost``, boost ``status``. Ditto for ``Favourite``.
            When it is a list/tuple, recursively call ``self._respond(status, n)``
            for each ``n`` in content.
        """
        if not content:
            raise ValueError(f"Response to {status['id']} empty; aborted")

        if type(content) in (list, tuple):
            for n in content:
                self._respond(status, n)
            return
        elif content == Boost:
            self._bot.status_reblog(status["id"])
            return
        elif content == Favourite:
            self._bot.status_favourite(status["id"])
            return
        elif type(content) == str:
            args = {
                STATUS: content,
                VISIBILITY: status[VISIBILITY],
                SPOILER_TEXT: status[SPOILER_TEXT],
            }
        elif isinstance(content, Reply):
            args = {
                STATUS: content.text,
                VISIBILITY: content.visibility or status[VISIBILITY],
                SPOILER_TEXT: content.spoiler_text or status[SPOILER_TEXT],
            }

        self._bot.status_post(in_reply_to_id=status["id"], **args)

    # decorator generators

    def on_mention(self, expectation, validation=EQUALS, case_sensitive=False):
        """Listen to mentions and invoke a callback with a ``Status`` object
            as argument.

        :param expectation: (str or callable) string, regex string or callable
            that evaluates to True if the status content is what you want.
        :param validation: (str) may be "equals", "contains", "regex" or "evaluate".
        """

        def decorator(callback):
            self._triggers.append(
                Trigger(
                    event=MENTION,
                    validation=validation,
                    expectation=expectation,
                    callback=callback,
                    case_sensitive=case_sensitive,
                )
            )

        return decorator

    def on_home_update(self, expectation, validation=EQUALS, case_sensitive=False):
        """Listen to updates on the home timeline and invoke a callback with
            a ``Status`` object as argument.

        :param expectation: (str or callable) string, regex string or callable
            that evaluates to True if the status content is what you want.
        :param validation: (str) may be "equals", "contains", "regex" or "evaluate".
        """

        def decorator(callback):
            self._triggers.append(
                Trigger(
                    event=UPDATE,
                    validation=validation,
                    expectation=expectation,
                    callback=callback,
                    case_sensitive=case_sensitive,
                )
            )

        return decorator

    # execution

    def run(self):
        """Start bot.

        After all listeners (triggers) are set, invoke ``bot.run()``.
        """
        self._bot = Mastodon(api_base_url=self._instance, access_token=self._token)
        print("Connected to " + self._instance)

        # keep a record of what this bot is called
        # so that "@atname" can be removed when necessary.
        me = self._bot.account_verify_credentials()
        self._handle = me["acct"]
        self._atname = "@" + me["username"]

        # register stream listeners
        if not self._websocket_mode:
            # use Mastodon.py StreamListener
            self._user_stream = StreamListener()
            self._user_stream.on_update = self._check_update_triggers
            self._user_stream.on_notification = self._check_notification_triggers
            self._bot.stream_user(self._user_stream)
        else:
            # use ws interface I implemented myself
            self._user_stream = WebsocketListener(
                self._bot.instance(), self._token, stream="user"
            )
            self._user_stream.on_update = self._check_update_triggers
            self._user_stream.on_notification = self._check_notification_triggers
            self._user_stream.start_stream()



from mastobot import *
from os import environ

if __name__ == "__main__":
    instance = environ.get("INSTANCE")
    token = environ.get("ACCESS_TOKEN")
    bot = Bot(instance, token, websocket_mode=True)

    """Test mention

    input: @bot Henlo
    output: @you hi, <your_display_name>
    """

    @bot.on_mention("henlo", validation=EQUALS, case_sensitive=False)
    def respond(status):
        return "hi, " + status.account.display_name

    """Test 'regex' validation

    input: :hacker_e: :hacker_d: :hacker_g: :hacker_y:
    outpur: nice hacker font you have
    """

    @bot.on_home_update("(:hacker_[a-z]:\s)+", validation=REGEX)
    def nice_hacker_font(status):
        return Reply("nice hacker font you have", visibility=UNLISTED)

    """Test 'evaluate' validation

    input: :cate: :cate: :cate:
    output: Woah, 3 cates - that's a lot!
    """

    has_more_than_two_cates = lambda s: s.count(":cate:") > 2

    @bot.on_home_update(has_more_than_two_cates, validation=EVALUATE)
    def so_many_cates(status):
        cate_count = status.text.count(":cate:")
        return Reply(f"Woah, {cate_count} cates - that's a lot!")

    """Test home timeline with 'contains' validation

    input: hey mastobot!
    output: yes i am online why do you ask
    """

    @bot.on_home_update("mastobot!", validation=CONTAINS)
    def online(status):
        return "yes i am online why do you ask"

    """Test 'equals' validation with list containing boost and fav as response

    input: mastobot plz favboost my post
    output: yes honey [fav] [boost]
    """

    @bot.on_home_update("mastobot plz favboost my post")
    def yes_honey(status):
        return ["yes honey", Favourite, Boost]

    """Test responding with nothing at all

    input: ksdfjdjgsakhghkalvfh
    output: [nothing at all]
    """

    @bot.on_home_update("ksdfjdjgsakhghkalvfh")
    def do_nothing(status):
        pass

    bot.run()
